#include <stdint.h> // for uint8_t
#include <sys/socket.h> // for recv...
#include <stdlib.h>     // for exit...
#include <sys/file.h>
#include <unistd.h>
#include <termios.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <pthread.h>

#include "util.h"
#include "debug.h"
#include "modem_core.h"
#include "ip.h"
#include "ip232.h"

void *ip232_thread(void *arg)
{
  modem_config *cfg = (modem_config *) arg;
  int accept_pending = FALSE;
  int rc;
  uint8_t buf[256];

  fd_set readfs;
  int max_fd = 0;
  int cSocket;


  LOG_ENTER();
  for (;;) {
    FD_ZERO(&readfs);
    FD_SET(cfg->dce_data.dp[1][0], &readfs);
    max_fd = cfg->dce_data.dp[1][0];
    if (accept_pending == FALSE) {
      FD_SET(cfg->dce_data.sSocket, &readfs);
      max_fd = MAX(max_fd, cfg->dce_data.sSocket);
    }
    LOG(LOG_DEBUG, "Waiting for incoming ip232 connections");
    rc = select(max_fd + 1, &readfs, NULL, NULL, NULL);
    if (rc < 0) { 
//      LOG(LOG_ALL, "ip232 select() returned an error: [%d] [%s]\n", errno, strerror(errno));
        ELOG(LOG_WARN, "IP232: Select returned error");
      // handle error
    } else {
      if (FD_ISSET(cfg->dce_data.dp[1][0], &readfs)) {  // pipe
        read(cfg->dce_data.dp[1][0], buf, sizeof(buf) - 1);
        LOG(LOG_DEBUG, "ip232 thread notified - Accept_pending set to false.");
        accept_pending = FALSE;
      }
      if (FD_ISSET(cfg->dce_data.sSocket, &readfs)) {   // ip connection
        if (cfg->dce_data.ip232_is_connected) {
          LOG(LOG_WARN, "Already have ip232 connection, rejecting new");

          // already have a connection... accept and close
          cSocket = ip_accept(cfg->dce_data.sSocket);
          if (cSocket > -1) {
		  // Send a message to Vice?
		  char *busy = "tcpser is already connected!\r\n";
		  int res =  send(cSocket, busy, strlen(busy), 0); // No need to check this for result, it's a courtesy and we are closing on the next line anyway. 
                  LOG(LOG_DEBUG, "ip232_thread: Sent Sockfd: [%d] Data: [%.*s] Len: [%d] Result: [%d]", cfg->dce_data.fd, (int) strlen(busy), busy, (int)strlen(busy), res); 
	    LOG(LOG_INFO, "%s(NOTICE): Closing Sockfd [%d]", __func__, cSocket); 
            close(cSocket);
          }
        } else {
          LOG(LOG_INFO, "Incoming ip232 connection");
          writePipe(cfg->dce_data.dp[0][1], MSG_ACCEPT);
          accept_pending = TRUE;
        }
      }
    }
  }
  LOG_EXIT();
}

int spawn_ip232_thread(modem_config *cfg)
{
  int rc;
  pthread_t thread_id;


  rc = pthread_create(&thread_id, NULL, ip232_thread, (void *) cfg);
  LOG(LOG_ALL, "ip232 thread ID=%d", (int) thread_id);
  if (rc < 0) {
    ELOG(LOG_FATAL, "ip232 thread could not be started");
    exit(-1);
  }
  return 0;
}

int ip232_init_conn(modem_config *cfg)
{
  int rc = -1;
  int port;


  LOG_ENTER();
  LOG(LOG_INFO, "Opening ip232 device");
  port = atoi(cfg->dce_data.tty);
  rc = ip_init_server_conn(NULL, port);
  if (rc < 0) {
    ELOG(LOG_FATAL, "Could not initialize ip232 server socket");
    exit(-1);
  }
  if (-1 == pipe(cfg->dce_data.dp[0])) {
    ELOG(LOG_FATAL, "ip232 thread incoming IPC pipe could not be created");
    exit(-1);
  }
  if (-1 == pipe(cfg->dce_data.dp[1])) {
    ELOG(LOG_FATAL, "ip232 thread outgoing IPC pipe could not be created");
    exit(-1);
  }
  cfg->dce_data.sSocket = rc;
  cfg->dce_data.ip232_is_connected = FALSE;
  cfg->dce_data.fd = cfg->dce_data.dp[0][0];
  spawn_ip232_thread(cfg);
  LOG(LOG_INFO, "ip232 device configured");
  LOG_EXIT();
  return rc;
}

int ip232_set_flow_control(modem_config *cfg, int status)
{
  return 0;
}

int ip232_get_control_lines(modem_config *cfg)
{
  int status = 0;


  if (cfg->dce_data.ip232_is_connected && cfg->dce_data.ip232_dtr) {
    status |= TIOCM_DSR;
  }
  return status;
}

int ip232_set_control_lines(modem_config *cfg, int state)
{
  int dcd;
  unsigned char cmd[2];


  if (cfg->dce_data.ip232_is_connected) {
    dcd = (state & TIOCM_DTR) ? TRUE : FALSE;
    if (dcd != cfg->dce_data.ip232_dcd) {
      cfg->dce_data.ip232_dcd = dcd;
      cmd[0] = 0xff;
      cmd[1] = dcd ? 1 : 0;
      ip_write(cfg->dce_data.fd, cmd, sizeof(cmd));
    }
  }
  return 0;
}

int ip232_write(modem_config *cfg, uint8_t *data, int len)
{
  int retval;
  int i = 0;
  int double_iac = FALSE;
  uint8_t text[1500];
  int text_len = 0;


  log_trace(TRACE_MODEM_OUT, (char *) data, len);
  retval = len;
  if (cfg->dce_data.ip232_is_connected) {
    while (i < len) {
      if (double_iac) {
        text[text_len++] = 0xff;
        double_iac = FALSE;
        i++;
      }
      else {
        if (0xff == data[i]) {
          text[text_len++] = 0xff;
          double_iac = TRUE;
        }
        else {
          text[text_len++] = data[i++];
        }
      }
      if (text_len == sizeof(text)) {
        retval = ip_write(cfg->dce_data.fd, text, text_len);
        text_len = 0;
      }
    }
    if (text_len) {
      retval = ip_write(cfg->dce_data.fd, text, text_len);
    }
  }
  return retval;
}

int ip232_read(modem_config *cfg, uint8_t *data, int len)
{
  int res;
  int rc;
  uint8_t buf[256];
  int i = 0;
  uint8_t ch;
  int text_len = 0;


  LOG_ENTER();
  if (len > sizeof(buf)) {
    LOG(LOG_FATAL, "ip232_read: len > sizeof(buf)");
    exit(-1);
  }
  if (cfg->dce_data.ip232_is_connected) {
    res = recv(cfg->dce_data.fd, buf, len, 0);
    if (0 >= res) {
      LOG(LOG_INFO, "No ip232 socket data read, assume closed peer");
      ip_disconnect(cfg->dce_data.fd);
      printf("%s(DEBUG): dce_data.fd[%d] was just closed - setting it to [%d]\n", __func__, cfg->dce_data.fd, cfg->dce_data.dp[0][0]);
      cfg->dce_data.fd = cfg->dce_data.dp[0][0]; // What is this doing?! Setting the dce sockfd to that of the pipe? 
      // cfg->dce_data.fd = -1;
      cfg->dce_data.ip232_is_connected = FALSE;
      // We should set DTR (?) here low perhaps? 
      cfg->dce_data.ip232_dtr = FALSE;
      cfg->dce_data.ip232_dcd = FALSE;
      LOG(LOG_DEBUG, "       --- DTR & DCD low due to connection loss with ip232 host (Vice)");

      writePipe(cfg->dce_data.dp[1][1], MSG_ACCEPT);
    }
    else {
//      LOG(LOG_DEBUG, "Read %d bytes from ip232 socket", res);
      log_trace(TRACE_MODEM_IN, (char*) buf, res);
      while (i < res) {
        ch = buf[i];
      LOG(LOG_DEBUG, "ip232_read: Sockfd: [%d] Data: [%02x] [%c]", cfg->dce_data.fd, ch, ch); 

	if (cfg->dce_data.ip232_iac) {
          cfg->dce_data.ip232_iac = FALSE;
          switch (ch) {
          case 0x00:
            cfg->dce_data.ip232_dtr = FALSE;
	    LOG(LOG_DEBUG, "       --- DTR is now low");
            break;
          case 0x01:
            cfg->dce_data.ip232_dtr = TRUE;
	    LOG(LOG_DEBUG, "       *** DTR is now HIGH");
            break;
          case 0xff:
	    LOG(LOG_DEBUG, "       /// Received an actual 0xff byte -- passing."); 
            data[text_len++] = 0xff;
            break;
          }
        }
        else {
          if (0xff == ch) {
            cfg->dce_data.ip232_iac = TRUE;
	    LOG(LOG_DEBUG, "ip232_read: GOT a potential Magic Byte, setting IAC"); 
          }
          else {
            data[text_len++] = ch;
          }
        }
        i++;
      }
    }
  }
  else {
    // not connected
    res = read(cfg->dce_data.dp[0][0], buf, sizeof(buf));
    switch (buf[0]) {
    case MSG_ACCEPT:   // accept connection.
      LOG(LOG_INFO, "Accepting ip232 connection...");
      rc = ip_accept(cfg->dce_data.sSocket);
      if (res > -1) {
        cfg->dce_data.fd = rc;
        cfg->dce_data.ip232_is_connected = TRUE;
        cfg->dce_data.ip232_dtr = FALSE;
        cfg->dce_data.ip232_dcd = FALSE;
        writePipe(cfg->dce_data.dp[1][1], MSG_ACCEPTED);
      }
      break;
    default: 
      LOG(LOG_INFO, "IP232: Received unknown data on pipe: [%c]", buf[0]); 
    }
  }

  LOG_EXIT();
  return text_len;
}
